/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: tun_tap_linux.cpp 2193 2017-04-02 00:42:59Z stephane $
 * See license.txt.
 */

#ifdef EPS_LINUX

#include "tun_tap.hpp"
#include "logging.hpp"
#include "common.hpp"
#include <boost/regex.hpp>
#include <linux/if_tun.h>
#include <linux/sockios.h>
#include <net/route.h>
#include <iomanip>
#include <fstream>
#include <cstdlib>
#include <cstring>
#include <string>


bool EPS::validate_network_against_existing_routes(const uint32_t &eps_network_in_host_order, const uint32_t &eps_netmask_in_host_order)
{
	/* Knowing the IP address and netmask assigned to us by the server, check to see if this will cause a problem.
	 * For example, if we've been assigned 192.168.1.123/24, but we already have an interface on a 192.168.1.x/24
	 * network, then bringing up the eps0 interface will cause problems.
	 */

	/* Match lines like these ones:
	 *
	 *      Iface   Destination Gateway     Flags   RefCnt  Use Metric  Mask        MTU Window  IRTT
	 *      eth0    00000000    0101A8C0    0003    0       0   0       00000000    0   0       0
	 *      eth0    0001A8C0    00000000    0001    0       0   1       00FFFFFF    0   0       0
	 */
	const boost::regex rx(
			"^"						// start of line
			"(\\w+)\\s+"			// interface name	-- group #1
			"([\\dA-Fa-f]{8})\\s+"	// destination		-- group #2
			"([\\dA-Fa-f]{8})\\s+"	// gateway			-- group #3
			"\\d+\\s+"				// flags
			"\\d+\\s+"				// RefCnt
			"\\d+\\s+"				// Use
			"(\\d+)\\s+"			// metric			-- group #4
			"([\\dA-Fa-f]{8})\\s+"	// netmask			-- group #5
			"(\\d+)\\s+"			// mtu				-- group #6
			".*"					// remaining (ignored) fields
			"$"						// end of line
			);

	const EPS::IPv4 eps_network(eps_network_in_host_order);

	size_t count = 0;
	std::string line;
	std::ifstream file("/proc/net/route");
	while (std::getline(file, line))
	{
		// must match the *entire* string to be valid
		boost::smatch what;
		if (not boost::regex_match(line, what, rx))
		{
			// line doesn't match
			continue;
		}

		count ++;

		const uint32_t dst_in_host_order		= ntohl(strtoul(what[2].str().c_str(), nullptr, 16));
		const uint32_t gateway_in_host_order	= ntohl(strtoul(what[3].str().c_str(), nullptr, 16));
		const uint32_t netmask_in_host_order	= ntohl(strtoul(what[5].str().c_str(), nullptr, 16));

		const std::string device(what[1]);
		const EPS::IPv4 dst	(dst_in_host_order);
		const EPS::IPv4 gw	(gateway_in_host_order);
		const EPS::IPv4 nm	(netmask_in_host_order);

		LOG_TRACE << "existing route #" << count << ": if=" << device << " dst=" << dst << " gw=" << gw << " nm=" << nm;

		const uint32_t netmask			= (netmask_in_host_order ? netmask_in_host_order : eps_netmask_in_host_order);
		const uint32_t dst_and_netmask	= (dst_in_host_order & netmask);
		const uint32_t gw_and_netmask	= (gateway_in_host_order & netmask);
		const uint32_t eps_and_netmask	= (eps_network_in_host_order & netmask);

		if (dst_and_netmask == eps_and_netmask)
		{
			LOG_ERROR << device << ": destination " << dst << " conflicts with the EPS network address " << eps_network;
			/// @return FALSE if a destination address matches the %EPS network address
			return false;
		}
		if (gw_and_netmask == eps_and_netmask)
		{
			LOG_ERROR << device << ": gateway " << gw << " conflicts with the EPS network address " << eps_network;
			/// @return FALSE if a gateway address matches the %EPS network address
			return false;
		}
	}

	/// @return TRUE if there doesn't seem to be a conflict
	return true;
}


void EPS::set_tun_tap_routes(
	const uint32_t		&network_in_network_order	,
	const uint32_t		&netmask_in_network_order	,
	const std::string	&device_name				)
{
	// this will add a temporary route which will automatically go away when the eps0 device is taken down

#if 0 // looks like this is unnecessary until we have to tweak some values

	rtentry route;
	memset(&route, '\0', sizeof(route));

	route.rt_metric	= 3;
	route.rt_flags	= RTF_UP | RTF_STATIC;
	route.rt_dev	= (char*)device_name.c_str();

	sockaddr_in *destination	= reinterpret_cast<sockaddr_in*>(&route.rt_dst);		// target address
	sockaddr_in *netmask		= reinterpret_cast<sockaddr_in*>(&route.rt_genmask);	// netmask
	sockaddr_in *gateway		= reinterpret_cast<sockaddr_in*>(&route.rt_gateway);	// gateway address

	destination	->sin_family		= AF_INET;
	netmask		->sin_family		= AF_INET;
	gateway		->sin_family		= AF_INET;

	destination	->sin_addr.s_addr	= network_in_network_order;
	netmask		->sin_addr.s_addr	= netmask_in_network_order;
	gateway		->sin_addr.s_addr	= INADDR_ANY;

	LOG_INFO	<< "adding route: "
				<< "destination: "	<< EPS::IPv4(ntohl(destination	->sin_addr.s_addr)).to_string()
				<< ", netmask: "	<< EPS::IPv4(ntohl(netmask		->sin_addr.s_addr)).to_string()
				<< ", gateway: "	<< EPS::IPv4(ntohl(gateway		->sin_addr.s_addr)).to_string()
				<< ", device: "		<< route.rt_dev;

	int sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
	int err = ioctl(sock, SIOCADDRT, &route);
	const int saved_errno = errno;
	close(sock);
	if (err < 0)
	{
		THROW_RUNTIME_ERROR(saved_errno, "failed to add gateway route");
	}

#endif

	return;
}


void EPS::raise_tun_tap_interface(const uint32_t ip_address_in_network_order, std::string &device_name, FILE_HANDLE_DESCRIPTOR &device_fd)
{
	device_name	= "";
	device_fd	= -1;

	// several web pages describing how to setup the tun/tap interface and ioctl calls:
	//
	// https://www.kernel.org/doc/Documentation/networking/tuntap.txt
	// http://vtun.cvs.sourceforge.net/viewvc/vtun/vtun/linux/tun_dev.c
	// http://backreference.org/2010/03/26/tuntap-interface-tutorial/
	// http://linuxgazette.net/149/melinte.html

	// open the clone device
	device_fd = open("/dev/net/tun", O_RDWR);
	if (device_fd < 0)
	{
		THROW_RUNTIME_ERROR(errno, "failed to open /dev/net/tun");
	}

	ifreq ifr_tun;
	memset(&ifr_tun, '\0', sizeof(ifr_tun));
	strncpy(ifr_tun.ifr_name, "eps%d", IFNAMSIZ);
	ifr_tun.ifr_flags = 0
//			| IFF_TUN		// TUN device, layer 3, IP packets, etc.
			| IFF_TAP		// TAP device, layer 2, Ethernet frames
			| IFF_NO_PI		// do not send extra packet information
			;

	int err = ioctl(device_fd, TUNSETIFF, &ifr_tun);
	int saved_errno = errno;
	if (err < 0)
	{
		if (saved_errno == EPERM)
		{
			THROW_RUNTIME_ERROR(saved_errno, "permission denied while setting up TUN/TAP interface (need to run as root?)");
		}
		THROW_RUNTIME_ERROR(saved_errno, "failed to setup TUN/TAP interface: err=" + std::to_string(err));
	}

	device_name = ifr_tun.ifr_name;
	LOG_INFO << "TUN/TAP device " << device_name << " has been created";

	// now we need to configure it (ip address, bring it up)
	//
	// http://stackoverflow.com/questions/5308090/set-ip-address-using-siocsifaddr-ioctl

	// If we throw we wont close the socket...but then we have bigger issues than a
	// leaked socket and the application will shutdown after the throw anyway...!
	int sock = socket(AF_INET, SOCK_DGRAM, IPPROTO_IP);
	if (sock < 0)
	{
		THROW_RUNTIME_ERROR(errno, "failed to create a new socket to configure the TUN/TAP interface");
	}

	sockaddr_in *addr		= reinterpret_cast<sockaddr_in*>(&ifr_tun.ifr_addr);
	addr->sin_addr.s_addr	= ip_address_in_network_order;
	addr->sin_family		= AF_INET;

	err = ioctl(sock, SIOCSIFADDR, &ifr_tun);
	if (err < 0)
	{
		THROW_RUNTIME_ERROR(errno, "failed to set the address for the TUN/TAP interface: err=" + std::to_string(err));
	}

	// get and then set the flags to bring the interface up
	ioctl(sock, SIOCGIFFLAGS, &ifr_tun);
	ifr_tun.ifr_flags |= IFF_UP;
	ifr_tun.ifr_flags |= IFF_RUNNING;
	err = ioctl(sock, SIOCSIFFLAGS, &ifr_tun);
	if (err < 0)
	{
		THROW_RUNTIME_ERROR(errno, "failed to set the flags on the TUN/TAP interface: err=" + std::to_string(err));
	}

	LOG_INFO << "TUN/TAP device " << ifr_tun.ifr_name << " (fd=" << device_fd << ") has been set to " << inet_ntoa(addr->sin_addr);

#if 0
	ifr_tun.ifr_mtu = 1400; /// @todo do we need to tweak the MTU?
	err = ioctl(sock, SIOCSIFMTU, &ifr_tun);
	if (err < 0)
	{
		// this is not fatal -- log it, but don't abort
		LOG_ERROR << "failed to set the mtu on the TUN/TAP interface: " << strerror(errno);
	}

	LOG_INFO << "TUN/TAP device " << ifr_tun.ifr_name << " mtu has been changed to " << ifr_tun.ifr_mtu;
#endif

	close(sock);

	return;
}


void EPS::close_tun_tap_interface(std::string &device_name, FILE_HANDLE_DESCRIPTOR &device_fd)
{
	// nothing specific to Linux to do here
	return;
}


#endif // EPS_LINUX
