/* EPS Conduits (C) 2014-2017 Stephane Charette <stephanecharette@gmail.com>
 * $Id: ServerHull.cpp 2195 2017-04-02 05:23:59Z stephane $
 * See license.txt.
 */

#include "ServerHull.hpp"
#include "logging.hpp"
#include "comm.hpp"
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/uuid/uuid_io.hpp>
#include <boost/uuid/string_generator.hpp>
#include <algorithm>
#include <thread>
#include <chrono>
#include <iomanip>
#include <time.h>


EPS::ServerHull::ServerHull(int argc, char *argv[]) :
	Hull(EPS::kServer, argc, argv),
	graphviz_network_diagram_timer(io_service)
{
	LOG_DEBUG << "outer server hull constructed";

	return;
}


EPS::ServerHull::~ServerHull(void)
{
	LOG_DEBUG << "outer server hull destruction in progress";
	this->done = true;

	graphviz_network_diagram_timer.cancel(this->ec);

	if (not this->tracker.empty())
	{
		this->tracker.export_to_cache(this->network);
	}

	#ifndef EPS_SKIP_UPNP
		reset_UPnP();
	#endif

	return;
}


void EPS::ServerHull::run_second_stage(void)
{
	while ( ! done )
	{
		auto until = std::chrono::system_clock::now() + seconds_between_retries;

		#ifndef EPS_SKIP_UPNP
			this->setup_UPnP();
		#endif

		this->tracker.import_from_cache(this->network);
		this->setup_address_pool();
		this->setup_listener();
		this->start_graphviz_network_diagram_timer();

		if (not this->ec)
		{
			LOG_INFO << "running IO loop";
			this->io_service.run(this->ec);
		}

		if (this->ec)
		{
			LOG_ERROR << "EPS server has failed: " << this->ec.message();
			this->ec.clear();
		}
		if ( ! done )
		{
			// prevent busy loops
			std::this_thread::sleep_until(until);
		}
	}

	return;
}


void EPS::ServerHull::setup_address_pool(void)
{
	this->ipv4_pool.clear();

	/// @todo this will need to be changed once we support anything other than a simple /24

	// remember to skip .0 and .255
	const unsigned long start	= this->network.to_ulong() + 1;		// 192.168.43.1
	const unsigned long end		= start + 253;						// 192.168.43.254

	// put the smallest address at the top of the vector (first to be popped)
	for ( unsigned long idx = end; idx >= start; idx --)
	{
		this->ipv4_pool.push_back( EPS::IPv4(idx) );
	}

	if (not this->sequential)
	{
		/* We don't want to rely on the 1st connection always being .1, second connection being .2, etc., because since
		 * we retain configuration and state information between restarts, the addresses aren't always that predictable.
		 * To ensure we don't have people accidentally relying on the order in which addresses are assigned, we may as
		 * well randomize the process.
		*/
		std::random_shuffle(ipv4_pool.begin(), ipv4_pool.end());
	}

	LOG_DEBUG << "IPv4 pool contains " << this->ipv4_pool.size() << " entries: " << IPv4(start).to_string() << " to " << IPv4(end).to_string();

	return;
}


void EPS::ServerHull::start_graphviz_network_diagram_timer(void)
{
	if (this->gv_filename.empty() == false)
	{
		#ifdef NDEBUG
			const int timer_expiry_in_minutes = 5;
		#else
			const int timer_expiry_in_minutes = 1;
		#endif

		this->graphviz_network_diagram_timer.expires_from_now(std::chrono::minutes(timer_expiry_in_minutes));
		this->graphviz_network_diagram_timer.async_wait(
				[=](const EC &error)
				{
					if (not error && not this->done)
					{
						generate_network_diagram();

						// schedule the next timer trigger
						this->start_graphviz_network_diagram_timer();
					}

					return;
				} );
	}

	return;
}


std::string gv_uuid(const EPS::UUID &uuid)
{
	// convert a UUID to a Graphviz-compatible ID string

	std::string str = "uuid_" + boost::uuids::to_string(uuid);

	// replace invalid characters with "_"
	const std::string valid_chars =
			"abcdefghijklmnopqrstuvwxyz"
			"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
			"_0123456789";
	size_t pos = 0;
	while (true)
	{
		pos = str.find_first_not_of(valid_chars, pos);
		if (pos == std::string::npos)
		{
			break;
		}

		str[pos] = '_';
	}

	return str;
}


std::string gv_colour(const EPS::UUID &uuid)
{
	const std::string u = boost::uuids::to_string(uuid);
	//
	// eg:  020a6b9e-3311-4c19-84b6-9f2378d758f9
	//								\rr/\gg/\bb/
	//
	//	red		= 0xff & (0x9f + 0x23)
	//	green	= 0xff & (0x78 + 0xd7)
	//	blue	= 0xff & (0x58 + 0xf9)

	const size_t pos = 24;

	unsigned long ul = 0;
	ul += strtoul(u.substr(pos + 0x00, 2).c_str(), nullptr, 16);
	ul += strtoul(u.substr(pos + 0x02, 2).c_str(), nullptr, 16);
	int red = ul & 0xff;

	ul = 0;
	ul += strtoul(u.substr(pos + 0x04, 2).c_str(), nullptr, 16);
	ul += strtoul(u.substr(pos + 0x06, 2).c_str(), nullptr, 16);
	int green = ul & 0xff;

	ul = 0;
	ul += strtoul(u.substr(pos + 0x08, 2).c_str(), nullptr, 16);
	ul += strtoul(u.substr(pos + 0x0A, 2).c_str(), nullptr, 16);
	int blue = ul & 0xff;

	// lighten the colour if it is too dark
	if (red		< 0x80) red		+= 0x80;
	if (green	< 0x80) green	+= 0x80;
	if (blue	< 0x80) blue	+= 0x80;

	std::stringstream ss;
	ss << "#"	<< std::hex << std::setw(2) << std::setfill('0') << red
				<< std::hex << std::setw(2) << std::setfill('0') << green
				<< std::hex << std::setw(2) << std::setfill('0') << blue;

	return ss.str();
}


void EPS::ServerHull::generate_network_diagram(void)
{
	LOG_INFO << "generating a new graphviz network diagram";

	const std::string s1 = "EPS network diagram for " + this->network_str;
	const std::string s2 = "Generated by " EPS_SYSTEM " EPS server v" EPS_VERSION;
	const std::string s3 = EPS::get_utc_timestamp();

	const std::string server_uuid = gv_uuid(this->uuid);

	// if the stats data is more than an hour old, don't display the details
	const time_t time_limit = std::time(nullptr) - (60*60);

	std::stringstream ss;
	ss	<< "# " << s1																				<< std::endl
		<< "# " << s2																				<< std::endl
		<< "# " << s3																				<< std::endl
		<< "#"																						<< std::endl
		<< "# PNG output:  circo -Tpng " << this->gv_filename << " > output.png"					<< std::endl
		<< "#"																						<< std::endl
		<< ""																						<< std::endl
		<< "digraph eps_network" 																	<< std::endl
		<< "{"																						<< std::endl
		<< "#	bgcolor=\"transparent\";"															<< std::endl
		<< "	bgcolor=\"white\";"																	<< std::endl
		<< "#	center=\"true\";"																	<< std::endl
		<< "	concentrate=\"true\";"																<< std::endl
		<< "	label=\"" + s1 + "\\n" + s2 + "\\n" + s3 + "\";"									<< std::endl
		<< "	labelloc=\"b\";"																	<< std::endl
		<< "	outputorder=\"edgesfirst\";"														<< std::endl
		<< "	overlap=\"false\";"																	<< std::endl
		<< "	overlap_shrink=\"true\";"															<< std::endl
		<< "	pack=\"true\";"																		<< std::endl
		<< "	root=\"" << server_uuid << "\";"													<< std::endl
		<< "	splines=\"curved\";"																<< std::endl
		<< ""																						<< std::endl
		<< "	# try changing the \"dpi\" as a simple way to generate a larger or smaller image"	<< std::endl
		<< "	dpi=\"72\";"																		<< std::endl
		<< ""																						<< std::endl
		<< "	node [shape=\"Mrecord\" style=\"filled\" fillcolor=\"#d0ffd0\" penwidth=\"2.0\"];"	<< std::endl
		<< "	edge [dir=\"both\" color=\"#802020\" penwidth=\"2.0\" weight=\"5.0\"];"				<< std::endl
		<< ""																						<< std::endl;

	char hostname[50];
	gethostname(hostname, sizeof(hostname));
	hostname[sizeof(hostname)-1] = '\0';

	const EPS::UUIDSet active_uuids = this->tracker.get_all_active_uuids();

	std::string label = "EPS SERVER";
	if (hostname[0] != '\0')
	{
		label += "|" + std::string(hostname) + "\\n" + EPS::get_os_name() + "\\nEPS v" + EPS_VERSION;
	}
	label += "\\nuptime: " + EPS::get_approximate_time(this->start_time) + " / " + EPS::get_approximate_time(EPS::get_boot_timestamp());
	label +=
			"|"
			"rx: "		+ EPS::format_bytes(this->stats.total_stats.server_bytes_received)	+ ": " + this->stats.total_stats.server_bps_in	+ "\\n"
			"tx: "		+ EPS::format_bytes(this->stats.total_stats.server_bytes_sent)		+ ": " + this->stats.total_stats.server_bps_out	+ "\\n"
			"fwd: "		+ EPS::format_bytes(this->stats.total_stats.server_bytes_forwarded)	+ ": " + this->stats.total_stats.server_bps_fwd + "\\n"
			"dropped: "	+ EPS::format_bytes(this->stats.total_stats.bytes_dropped) +
			"|"
			"client connections: " + std::to_string(active_uuids.size());

	ss	<< "\t# EPS server:"	<< std::endl
		<< ""					<< std::endl
		<< "\t" << server_uuid << " [label=\"{" << label << "}\" fillcolor=\"#ffd0d0\" root=\"true\"];" << std::endl;

	// connect the server to all active UUIDs
	for (const auto &client_uuid : active_uuids)
	{
		ss << "\t" << server_uuid << " -> " << gv_uuid(client_uuid) << ";" << std::endl;
	}

	// now display all of the clients

	ss	<< ""												<< std::endl
		<< "\tedge [color=\"#000000\" penwidth=\"1.0\" weight=\"1.0\"];"	<< std::endl
		<< ""												<< std::endl
		<< "\t# EPS clients:"								<< std::endl;

	for (const EPS::UUID &u : this->tracker.get_all_uuids())
	{
		const EPS::TrackerData &data = this->tracker.get(u);
		const std::string client_uuid = gv_uuid(data.uuid);

		if (data.hostname.empty() && data.other_addresses.empty())
		{
			// this entry is from the cache, and we know (nearly) nothing about it, so skip it completely
			continue;
		}

		const bool is_inactive = data.is_inactive();

		label.clear();
		if (is_inactive)
		{
			label += "DISCONNECTED: ";
		}
		label += data.hostname;
		if (not data.os_name.empty())
		{
			label += "\\n" + data.os_name;
		}
		if (not data.eps_version.empty())
		{
			label += "\\nEPS v" + data.eps_version;
			if (data.eps_version != EPS_VERSION)
			{
				label += " [!]";
			}
		}

		// do not display uptime if disconnected (numbers will be wrong)
		if (is_inactive)
		{
			if (data.disconnect_timestamp > 0)
			{
				label += "\\ndisconnected: " + EPS::get_approximate_time(data.disconnect_timestamp) + " ago";
			}
		}
		else
		{
			label += "\\nuptime: " + EPS::get_approximate_time(data.eps_uptime) + " / " + EPS::get_approximate_time(data.system_uptime);
		}

		label += "|";

		if (data.other_addresses.empty() == false)
		{
			label += "addr: " + ipset_to_string(data.other_addresses) + "\\n";
		}

		label += "eps: " + data.address_eps.to_string();

		EPS::MUuidStats::const_iterator iter = this->uuid_stats.find(data.uuid);
		if (iter != this->uuid_stats.end())
		{
			// we have stats for this client connection
			const EPS::CommStats1 &client_stats = iter->second;

			if (time_t(ntohl(client_stats.timestamp)) >= time_limit)
			{
				// the stats are relatively recent, so include them in the label name
				label +=
						"|"
						"server rx: "	+ EPS::format_bytes(client_stats.server_bytes_received)	+ ": " + client_stats.server_bps_in		+ "\\n"
						"server tx: "	+ EPS::format_bytes(client_stats.server_bytes_sent)		+ ": " + client_stats.server_bps_out	;

				if (client_stats.direct_bytes_received || client_stats.direct_bytes_sent)
				{
					label +=
						"|"
						"mesh rx: "	+ EPS::format_bytes(client_stats.direct_bytes_received)	+ ": " + client_stats.direct_bps_in		+ "\\n"
						"mesh tx: " + EPS::format_bytes(client_stats.direct_bytes_sent)		+ ": " + client_stats.direct_bps_out	;
				}
				label +=
						"|"
						"tap rx: "	+ EPS::format_bytes(client_stats.tap_bytes_received)	+ ": " + client_stats.tap_bps_in		+ "\\n"
						"tap tx: "	+ EPS::format_bytes(client_stats.tap_bytes_sent)		+ ": " + client_stats.tap_bps_out		;
				if (client_stats.mesh_connections)
				{
					label +=
						"|"
						"mesh connections: " + std::to_string(ntohl(client_stats.mesh_connections));
				}
			}

			label += "|updated " + EPS::get_utc_timestamp(time_t(ntohl(client_stats.timestamp)));
		}

		ss	<< "" << std::endl
			<< "\t" << client_uuid << " [label=\"{" << label << "}\" fillcolor=\"" << gv_colour(data.uuid) << "\"];" << std::endl;

		// now see if we have some peers (mesh) we can link up
		for (const auto &mesh_uuid : this->uuid_mesh[data.uuid])
		{
			ss << "\t" << client_uuid << " -> " << gv_uuid(mesh_uuid);
			if (data.is_inactive())
			{
				ss << " [style=\"dotted\" dir=\"none\" weight=\"0.0\"]";
			}
			ss << ";" << std::endl;
		}
	}

	ss	<< "}"	<< std::endl
		<< ""	<< std::endl;

	EPS::write_file(this->gv_filename, ss.str());

	return;
}


void EPS::ServerHull::connection_failed( const EPS::TConnPtr &connection )
{
	if (connection)
	{
		this->stats.minute_stats.connections_dropped ++;

		LOG_WARN << connection->name << "connection has failed and will be deleted";

		connection->socket.cancel(this->ec);
		if (this->tracker.exists(connection))
		{
			this->tracker.deactivate_entry(connection);
		}
	}

	return;
}


void EPS::ServerHull::processComm( const EPS::TConnPtr &connection, EPS::NetBuffer *nb, const EPS::CommHeader1 *header, const EPS::ECommType type )
{

//	LOG_TRACE << connection->name << "handling " << EPS::comm_type_to_string(type) << " packet";

	switch (type)
	{
		case EPS::kData1:
		{
			/** @todo If we only decrypted enough to read the full CommData1 header, then we wouldn't have to decrypt
			 * the entire buffer.  We'd have the destination to which the encrypted packet needs to be sent.  But at
			 * the moment we either decrypt just the common header or the entire packet.
			 */
			const EPS::CommData1 *	cd1					= reinterpret_cast<const EPS::CommData1*>(header);
			const size_t			number_of_bytes		= cd1->header.size_of_packet();
			const size_t			bytes_with_padding	= EPS::length_including_pkcs_padding(number_of_bytes);

			// do we know the intended recipient of this packet?
			EPS::UUID uuid;
			memcpy(uuid.data, cd1->dst_uuid, uuid.static_size());
			if (not uuid.is_nil() && this->tracker.exists(uuid))
			{
				// we know where this packet needs to be written

				EPS::TrackerData &data = this->tracker.get(uuid);
				if (data.is_inactive())
				{
					this->stats.minute_stats.bytes_dropped		+= bytes_with_padding;
					this->stats.minute_stats.packets_dropped	++;

					// if a EPS client has died, and someone is still trying to send packets to it,
					// this can result in a large number of messages, so we need a way to limit what we log
					static	size_t	packet_counter	= 0;
					static	time_t	timestamp		= 0;
					const	time_t	now				= std::time(nullptr);

					packet_counter ++;

					if (timestamp == 0)
					{
						timestamp = now;
					}
					else if (now > timestamp + 300 || packet_counter >= 1000)
					{
						LOG_WARN << packet_counter << " packets dropped in the last " << EPS::get_approximate_time(timestamp) << " due to inactive client state; e.g., from " << connection->name << " " << bytes_with_padding << "-byte " << EPS::proto_to_string(cd1) << " packet to " << data.as_debug();
						timestamp		= 0;
						packet_counter	= 0;
					}
				}
				else
				{
//					LOG_TRACE << connection->id << "=>" << data.connection->id << ": forwarding " << number_of_bytes << "-byte " << EPS::proto_to_string(cd1) << " packet to a single endpoint";

					EPS::NetBuffer *encrypted_buffer = new EPS::NetBuffer(*nb->get_encrypted());
					this->async_write(data.connection, encrypted_buffer);

					this->stats.minute_stats.server_bytes_forwarded		+= encrypted_buffer->length;
					this->stats.minute_stats.server_packets_forwarded	++;
				}
			}
			else
			{
				// forward this data pulse to *all* clients so they can re-inject it into the local network
//				LOG_TRACE << connection->name << "forwarding " << number_of_bytes << "-byte " << EPS::proto_to_string(cd1) << " packet to all endpoints";

				for (auto conn : this->tracker.get_all_connections_other_than(connection))
				{
//					LOG_TRACE << connection->id << "=>" << conn->id << ": forwarding " << number_of_bytes << "-byte " << EPS::proto_to_string(cd1) << " packet to many clients";

					EPS::NetBuffer *encrypted_buffer = new EPS::NetBuffer(*nb->get_encrypted());
					this->async_write(conn, encrypted_buffer);

					this->stats.minute_stats.server_bytes_forwarded		+= encrypted_buffer->length;
					this->stats.minute_stats.server_packets_forwarded	++;
				}
			}
			break;
		}
		case EPS::kStats1:
		{
			const EPS::CommStats1 *	cs1 = reinterpret_cast<const EPS::CommStats1*>(header);
			this->uuid_stats[connection->uuid] = *cs1;

			// clear out all previously known connections for this client
			UUIDSet &all_peer_uuids = this->uuid_mesh[connection->uuid];
			all_peer_uuids.clear();

			// rebuild the known mesh connections for this client
			for (size_t idx = 0; idx < ntohl(cs1->mesh_connections); idx ++)
			{
				EPS::UUID mesh_uuid;
				memcpy(mesh_uuid.data, cs1->uuid[idx], UUID::static_size());
				all_peer_uuids.insert(mesh_uuid);
			}

			LOG_TRACE << connection->name << "received statistics from client (mesh connections: " << ntohl(cs1->mesh_connections) << ", in: " << cs1->tap_bps_in << ", out: " << cs1->tap_bps_out << ")";

			break;
		}
		case EPS::kHello2:
		{
			// keep track of this new client

			const EPS::CommHello2 *ch2 = reinterpret_cast<const EPS::CommHello2*>(header);
			auto gen = boost::uuids::string_generator();
			const UUID uuid = gen(ch2->uuid_str);

			EPS::TrackerData &data = this->tracker.get(uuid);

			if (data.connection.get()	!= nullptr		&&
				data.connection			!= connection	)
			{
				/* Something is wrong:  this uuid already has a connection, but we're receiving a duplicate connection
				 * for the same uuid.  Assume the client died, that we haven't detected it yet, and has restarted.
				 */
				LOG_WARN << connection->name << "new connection from the same uuid, need to shut down the old connection to: " << data.as_debug();
				this->connection_failed(data.connection);
			}

			data.connection	= connection;
			data.mode		= static_cast<EPS::EMode>(ch2->mode);

			if (not data.address_eps.is_unspecified())
			{
				// looks like we may have an address to use for this uuid....but is that address still available,
				// or did some other active client get assigned the same address?

				bool duplicate_eps_address_found = false;
				EPS::UUIDSet all_uuids = this->tracker.get_all_active_uuids();
				for (const auto &other_uuid : all_uuids)
				{
					const EPS::TrackerData &other_data = this->tracker.get(other_uuid);
					if (uuid != other_uuid && other_data.address_eps == data.address_eps)
					{
						// oops, looks like we've assigned that address to another client
						LOG_TRACE << connection->name << "ip address " << data.address_eps << " has already been assigned to: " << other_data.as_debug();
						duplicate_eps_address_found = true;
						// blank out the address so we'll be forced to assign a new one
						data.address_eps = IPv4();
						break;
					}
				}

				if (duplicate_eps_address_found == false)
				{
					LOG_TRACE << connection->name << "re-using previous address " << data.address_eps << " for uuid " << uuid;

					// this address is still available...but remove it from the IP pool so it doesn't accidentally get re-assigned to someone else
					const IPv4Vec::iterator iter = std::find(this->ipv4_pool.begin(), this->ipv4_pool.end(), data.address_eps);
					if (iter != this->ipv4_pool.end())
					{
						LOG_TRACE << connection->name << "removing " << data.address_eps << " from the IP pool";
						this->ipv4_pool.erase(iter);
					}
				}
			}

			// if we don't yet have an address, see if we can find a convenient one to use
			if (data.address_eps.is_unspecified() && not data.other_addresses.empty() && not this->sequential)
			{
				// we have a local address and no EPS address, so see if we can re-use the last part of the address
				const IPv4 ip = *data.other_addresses.begin();
				const std::string	tmp					= ip.to_string();
				const size_t		pos					= tmp.rfind('.');
				const std::string	digits				= tmp.substr(pos+1);
				const uint32_t		possible_address	= this->network.to_ulong() + atoi(digits.c_str());

				// see if the address is available
				const IPv4 addr(possible_address);
				LOG_TRACE << connection->name << "looking to see if " << addr << " is available for " << ip;

				const IPv4Vec::iterator iter = std::find(this->ipv4_pool.begin(), this->ipv4_pool.end(), addr);
				if (iter != this->ipv4_pool.end())
				{
					data.address_eps = addr;
					LOG_TRACE << connection->name << "assigned " << addr << " (similar to " << ip << ")";
					this->ipv4_pool.erase(iter);
				}
			}

			// last chance -- if we haven't yet found an eps address to use, then we'll take one from the pool
			if (data.address_eps.is_unspecified())
			{
				IPv4 ipv4 = this->ipv4_pool.back();
				this->ipv4_pool.pop_back();
				data.address_eps = ipv4;
				LOG_TRACE << connection->name << "assigned " << ipv4 << " to uuid " << uuid << " (number of IPv4 addresses remaining in pool: " << this->ipv4_pool.size() << ")";
			}

			// send back an address structure to the client

			EPS::NetBuffer *outgoing = new EPS::NetBuffer(EPS::NetBuffer::EType::kFullyDecrypted);
			EPS::CommAddress1 *ca1 = reinterpret_cast<EPS::CommAddress1*>(outgoing->data());
			ca1->set(this->network, this->netmask, data.address_eps);
			outgoing->length = ca1->header.size_of_packet();

			EPS::NetBuffer *encrypted = outgoing->get_encrypted();
			this->async_write(connection, outgoing);

			this->stats.minute_stats.server_bytes_sent		+= encrypted->length;
			this->stats.minute_stats.server_packets_sent	++;

			// now that we've received a "hello" message, send an updated peer list to all connected clients

			// create a list of all clients
			// (since this is the server, we know all connections will be client connections)

			outgoing = new EPS::NetBuffer(EPS::NetBuffer::EType::kFullyDecrypted);
			EPS::CommList2 *cl2 = reinterpret_cast<EPS::CommList2*>(outgoing->data());

			size_t counter = 0;
			for (EPS::UUID uuid : this->tracker.get_all_active_uuids())
			{
				const EPS::TrackerData &tracker_data = this->tracker.get(uuid);

				std::copy(tracker_data.uuid.begin(), tracker_data.uuid.end(), cl2->data[counter].uuid);
				cl2->data[counter].address_eps = htonl(tracker_data.address_eps.to_ulong());

				// zero out all the addresses so unused ones don't show up as garbage!
				for (size_t idx = 0; idx < MAX_NUMBER_OF_OTHER_ADDRESSES; idx++)
				{
					cl2->data[counter].other_addresses[idx] = 0;
				}

				size_t idx = 0;
				for ( const auto &ip : tracker_data.other_addresses )
				{
					// don't tell other clients about 127.0.0.1 addresses (if the client and server are on the same box)
					if (ip.is_loopback())
					{
						continue;
					}

					if (idx >= MAX_NUMBER_OF_OTHER_ADDRESSES)
					{
						// no more room, stop trying to fit more addresses
						break;
					}

					cl2->data[counter].other_addresses[idx] = htonl(ip.to_ulong());
					idx ++;
				}

				strncpy(cl2->data[counter].hostname_str, tracker_data.hostname.data(), EPS_HOSTNAME_LENGTH);

				LOG_INFO << "list[" << counter << "]: " << tracker_data.uuid << ": " << ipset_to_string(tracker_data.other_addresses) << " aka " << tracker_data.address_eps << " [" << tracker_data.hostname << "]";

				counter ++;

				if (counter >= 20)
				{
					LOG_WARN << "too many entries to send all of them (" << this->tracker.get_all_active_uuids().size() << ")";
					break;
				}
			}
			cl2->set(counter);
			outgoing->length = cl2->header.size_of_packet();

			for (auto conn : this->tracker.get_all_connections())
			{
				LOG_TRACE << conn->name << "sending peer list: " << counter << " records";

				EPS::NetBuffer *encrypted_buffer = new EPS::NetBuffer(*outgoing->get_encrypted());
				this->async_write(conn, encrypted_buffer);

				this->stats.minute_stats.server_bytes_sent		+= encrypted_buffer->length;
				this->stats.minute_stats.server_packets_sent	++;
			}

			break;
		}
		default:
		{
			break;
		}
	}

	return;
}
